@@ -5,29 +5,29 @@
 use crate::export::Export;
 use crate::externref::VMExternRefActivationsTable;
 use crate::memory::{Memory, RuntimeMemoryCreator};
-use crate::table::{Table, TableElement};
+use crate::table::{Table, TableElement, TableElementType};
 use crate::traphandlers::Trap;
 use crate::vmcontext::{
     VMCallerCheckedAnyfunc, VMContext, VMFunctionImport, VMGlobalDefinition, VMGlobalImport,
     VMInterrupts, VMMemoryDefinition, VMMemoryImport, VMTableDefinition, VMTableImport,
 };
 use crate::{ExportFunction, ExportGlobal, ExportMemory, ExportTable, Store};
+use anyhow::Error;
 use memoffset::offset_of;
 use more_asserts::assert_lt;
 use std::alloc::Layout;
 use std::any::Any;
-use std::collections::HashMap;
 use std::convert::TryFrom;
 use std::hash::Hash;
+use std::ops::Range;
 use std::ptr::NonNull;
 use std::sync::Arc;
 use std::{mem, ptr, slice};
-use wasmtime_environ::entity::{packed_option::ReservedValue, EntityRef, EntitySet, PrimaryMap};
-use wasmtime_environ::wasm::{
-    DataIndex, DefinedGlobalIndex, DefinedMemoryIndex, DefinedTableIndex, ElemIndex, EntityIndex,
-    FuncIndex, GlobalIndex, MemoryIndex, TableElementType, TableIndex, WasmType,
+use wasmtime_environ::{
+    packed_option::ReservedValue, DataIndex, DefinedGlobalIndex, DefinedMemoryIndex,
+    DefinedTableIndex, ElemIndex, EntityIndex, EntityRef, EntitySet, FuncIndex, GlobalIndex,
+    HostPtr, MemoryIndex, Module, PrimaryMap, TableIndex, TrapCode, VMOffsets, WasmType,
 };
-use wasmtime_environ::{ir, HostPtr, Module, VMOffsets};
 
 mod allocator;
 
@@ -45,18 +45,33 @@ pub const DEFAULT_MEMORY_LIMIT: usize = 10000;
 /// An instance can be created with a resource limiter so that hosts can take into account
 /// non-WebAssembly resource usage to determine if a linear memory or table should grow.
 pub trait ResourceLimiter {
-    /// Notifies the resource limiter that an instance's linear memory has been requested to grow.
+    /// Notifies the resource limiter that an instance's linear memory has been
+    /// requested to grow.
     ///
-    /// * `current` is the current size of the linear memory in WebAssembly page units.
-    /// * `desired` is the desired size of the linear memory in WebAssembly page units.
-    /// * `maximum` is either the linear memory's maximum or a maximum from an instance allocator,
-    ///   also in WebAssembly page units. A value of `None` indicates that the linear memory is
-    ///   unbounded.
+    /// * `current` is the current size of the linear memory in bytes.
+    /// * `desired` is the desired size of the linear memory in bytes.
+    /// * `maximum` is either the linear memory's maximum or a maximum from an
+    ///   instance allocator, also in bytes. A value of `None`
+    ///   indicates that the linear memory is unbounded.
     ///
-    /// This function should return `true` to indicate that the growing operation is permitted or
-    /// `false` if not permitted. Returning `true` when a maximum has been exceeded will have no
-    /// effect as the linear memory will not grow.
-    fn memory_growing(&mut self, current: u32, desired: u32, maximum: Option<u32>) -> bool;
+    /// This function should return `true` to indicate that the growing
+    /// operation is permitted or `false` if not permitted. Returning `true`
+    /// when a maximum has been exceeded will have no effect as the linear
+    /// memory will not grow.
+    ///
+    /// This function is not guaranteed to be invoked for all requests to
+    /// `memory.grow`. Requests where the allocation requested size doesn't fit
+    /// in `usize` or exceeds the memory's listed maximum size may not invoke
+    /// this method.
+    fn memory_growing(&mut self, current: usize, desired: usize, maximum: Option<usize>) -> bool;
+
+    /// Notifies the resource limiter that growing a linear memory, permitted by
+    /// the `memory_growing` method, has failed.
+    ///
+    /// Reasons for failure include: the growth exceeds the `maximum` passed to
+    /// `memory_growing`, or the operating system failed to allocate additional
+    /// memory. In that case, `error` might be downcastable to a `std::io::Error`.
+    fn memory_grow_failed(&mut self, _error: &Error) {}
 
     /// Notifies the resource limiter that an instance's table has been requested to grow.
     ///
@@ -141,6 +156,12 @@ pub(crate) struct Instance {
     /// If the index is present in the set, the segment has been dropped.
     dropped_data: EntitySet<DataIndex>,
 
+    /// A slice pointing to all data that is referenced by this instance. This
+    /// data is managed externally so this is effectively an unsafe reference,
+    /// and this does not live for the `'static` lifetime so the API boundaries
+    /// here are careful to never hand out static references.
+    wasm_data: &'static [u8],
+
     /// Hosts can store arbitrary per-instance information here.
     ///
     /// Most of the time from Wasmtime this is `Box::new(())`, a noop
@@ -406,8 +427,9 @@ impl Instance {
     /// Grow memory by the specified amount of pages.
     ///
     /// Returns `None` if memory can't be grown by the specified amount
-    /// of pages.
-    pub(crate) fn memory_grow(&mut self, index: MemoryIndex, delta: u32) -> Option<u32> {
+    /// of pages. Returns `Some` with the old size in bytes if growth was
+    /// successful.
+    pub(crate) fn memory_grow(&mut self, index: MemoryIndex, delta: u64) -> Option<usize> {
         let (idx, instance) = if let Some(idx) = self.module.defined_memory_index(index) {
             (idx, self)
         } else {
@@ -502,22 +524,6 @@ impl Instance {
         self.vmctx_plus_offset(self.offsets.vmctx_anyfuncs_begin())
     }
 
-    fn find_passive_segment<'a, I, D, T>(
-        index: I,
-        index_map: &HashMap<I, usize>,
-        data: &'a Vec<D>,
-        dropped: &EntitySet<I>,
-    ) -> &'a [T]
-    where
-        D: AsRef<[T]>,
-        I: EntityRef + Hash,
-    {
-        match index_map.get(&index) {
-            Some(index) if !dropped.contains(I::new(*index)) => data[*index].as_ref(),
-            _ => &[],
-        }
-    }
-
     /// The `table.init` operation: initializes a portion of a table with a
     /// passive element.
     ///
@@ -537,12 +543,13 @@ impl Instance {
         // inform `rustc` that the lifetime of the elements here are
         // disconnected from the lifetime of `self`.
         let module = self.module.clone();
-        let elements = Self::find_passive_segment(
-            elem_index,
-            &module.passive_elements_map,
-            &module.passive_elements,
-            &self.dropped_elements,
-        );
+
+        let elements = match module.passive_elements_map.get(&elem_index) {
+            Some(index) if !self.dropped_elements.contains(elem_index) => {
+                module.passive_elements[*index].as_ref()
+            }
+            _ => &[],
+        };
         self.table_init_segment(table_index, elements, dst, src, len)
     }
 
@@ -563,7 +570,7 @@ impl Instance {
             .and_then(|s| s.get(..usize::try_from(len).unwrap()))
         {
             Some(elements) => elements,
-            None => return Err(Trap::wasm(ir::TrapCode::TableOutOfBounds)),
+            None => return Err(Trap::wasm(TrapCode::TableOutOfBounds)),
         };
 
         match table.element_type() {
@@ -582,7 +589,7 @@ impl Instance {
                 )?;
             },
 
-            TableElementType::Val(_) => {
+            TableElementType::Extern => {
                 debug_assert!(elements.iter().all(|e| *e == FuncIndex::reserved_value()));
                 table.fill(dst, TableElement::ExternRef(None), len)?;
             }
@@ -594,9 +601,7 @@ impl Instance {
     pub(crate) fn elem_drop(&mut self, elem_index: ElemIndex) {
         // https://webassembly.github.io/reference-types/core/exec/instructions.html#exec-elem-drop
 
-        if let Some(index) = self.module.passive_elements_map.get(&elem_index) {
-            self.dropped_elements.insert(ElemIndex::new(*index));
-        }
+        self.dropped_elements.insert(elem_index);
 
         // Note that we don't check that we actually removed a segment because
         // dropping a non-passive segment is a no-op (not a trap).
@@ -616,26 +621,18 @@ impl Instance {
     pub(crate) fn memory_copy(
         &mut self,
         dst_index: MemoryIndex,
-        dst: u32,
+        dst: u64,
         src_index: MemoryIndex,
-        src: u32,
-        len: u32,
+        src: u64,
+        len: u64,
     ) -> Result<(), Trap> {
         // https://webassembly.github.io/reference-types/core/exec/instructions.html#exec-memory-copy
 
         let src_mem = self.get_memory(src_index);
         let dst_mem = self.get_memory(dst_index);
 
-        if src.checked_add(len).map_or(true, |n| {
-            usize::try_from(n).unwrap() > src_mem.current_length
-        }) || dst.checked_add(len).map_or(true, |m| {
-            usize::try_from(m).unwrap() > dst_mem.current_length
-        }) {
-            return Err(Trap::wasm(ir::TrapCode::HeapOutOfBounds));
-        }
-
-        let dst = usize::try_from(dst).unwrap();
-        let src = usize::try_from(src).unwrap();
+        let src = self.validate_inbounds(src_mem.current_length, src, len)?;
+        let dst = self.validate_inbounds(dst_mem.current_length, dst, len)?;
 
         // Bounds and casts are checked above, by this point we know that
         // everything is safe.
@@ -648,6 +645,19 @@ impl Instance {
         Ok(())
     }
 
+    fn validate_inbounds(&self, max: usize, ptr: u64, len: u64) -> Result<usize, Trap> {
+        let oob = || Trap::wasm(TrapCode::HeapOutOfBounds);
+        let end = ptr
+            .checked_add(len)
+            .and_then(|i| usize::try_from(i).ok())
+            .ok_or_else(oob)?;
+        if end > max {
+            Err(oob())
+        } else {
+            Ok(ptr as usize)
+        }
+    }
+
     /// Perform the `memory.fill` operation on a locally defined memory.
     ///
     /// # Errors
@@ -656,25 +666,17 @@ impl Instance {
     pub(crate) fn memory_fill(
         &mut self,
         memory_index: MemoryIndex,
-        dst: u32,
-        val: u32,
-        len: u32,
+        dst: u64,
+        val: u8,
+        len: u64,
     ) -> Result<(), Trap> {
         let memory = self.get_memory(memory_index);
-
-        if dst.checked_add(len).map_or(true, |m| {
-            usize::try_from(m).unwrap() > memory.current_length
-        }) {
-            return Err(Trap::wasm(ir::TrapCode::HeapOutOfBounds));
-        }
-
-        let dst = isize::try_from(dst).unwrap();
-        let val = val as u8;
+        let dst = self.validate_inbounds(memory.current_length, dst, len)?;
 
         // Bounds and casts are checked above, by this point we know that
         // everything is safe.
         unsafe {
-            let dst = memory.base.offset(dst);
+            let dst = memory.base.add(dst);
             ptr::write_bytes(dst, val, len as usize);
         }
 
@@ -692,50 +694,42 @@ impl Instance {
         &mut self,
         memory_index: MemoryIndex,
         data_index: DataIndex,
-        dst: u32,
+        dst: u64,
         src: u32,
         len: u32,
     ) -> Result<(), Trap> {
-        // TODO: this `clone()` shouldn't be necessary but is used for now to
-        // inform `rustc` that the lifetime of the elements here are
-        // disconnected from the lifetime of `self`.
-        let module = self.module.clone();
-        let data = Self::find_passive_segment(
-            data_index,
-            &module.passive_data_map,
-            &module.passive_data,
-            &self.dropped_data,
-        );
-        self.memory_init_segment(memory_index, &data, dst, src, len)
+        let range = match self.module.passive_data_map.get(&data_index).cloned() {
+            Some(range) if !self.dropped_data.contains(data_index) => range,
+            _ => 0..0,
+        };
+        self.memory_init_segment(memory_index, range, dst, src, len)
+    }
+
+    pub(crate) fn wasm_data(&self, range: Range<u32>) -> &[u8] {
+        &self.wasm_data[range.start as usize..range.end as usize]
     }
 
     pub(crate) fn memory_init_segment(
         &mut self,
         memory_index: MemoryIndex,
-        data: &[u8],
-        dst: u32,
+        range: Range<u32>,
+        dst: u64,
         src: u32,
         len: u32,
     ) -> Result<(), Trap> {
         // https://webassembly.github.io/bulk-memory-operations/core/exec/instructions.html#exec-memory-init
 
         let memory = self.get_memory(memory_index);
+        let data = self.wasm_data(range);
+        let dst = self.validate_inbounds(memory.current_length, dst, len.into())?;
+        let src = self.validate_inbounds(data.len(), src.into(), len.into())?;
+        let len = len as usize;
 
-        if src
-            .checked_add(len)
-            .map_or(true, |n| usize::try_from(n).unwrap() > data.len())
-            || dst.checked_add(len).map_or(true, |m| {
-                usize::try_from(m).unwrap() > memory.current_length
-            })
-        {
-            return Err(Trap::wasm(ir::TrapCode::HeapOutOfBounds));
-        }
-
-        let src_slice = &data[src as usize..(src + len) as usize];
+        let src_slice = &data[src..(src + len)];
 
         unsafe {
-            let dst_start = memory.base.add(dst as usize);
-            let dst_slice = slice::from_raw_parts_mut(dst_start, len as usize);
+            let dst_start = memory.base.add(dst);
+            let dst_slice = slice::from_raw_parts_mut(dst_start, len);
             dst_slice.copy_from_slice(src_slice);
         }
 
@@ -744,9 +738,7 @@ impl Instance {
 
     /// Drop the given data segment, truncating its length to zero.
     pub(crate) fn data_drop(&mut self, data_index: DataIndex) {
-        if let Some(index) = self.module.passive_data_map.get(&data_index) {
-            self.dropped_data.insert(DataIndex::new(*index));
-        }
+        self.dropped_data.insert(data_index);
 
         // Note that we don't check that we actually removed a segment because
         // dropping a non-passive segment is a no-op (not a trap).
