@@ -4,7 +4,7 @@ use crate::memory::{DefaultMemoryCreator, Memory};
 use crate::table::Table;
 use crate::traphandlers::Trap;
 use crate::vmcontext::{
-    VMBuiltinFunctionsArray, VMCallerCheckedAnyfunc, VMContext, VMFunctionBody, VMGlobalDefinition,
+    VMBuiltinFunctionsArray, VMCallerCheckedAnyfunc, VMContext, VMGlobalDefinition,
     VMSharedSignatureIndex,
 };
 use crate::Store;
@@ -17,13 +17,10 @@ use std::ptr::{self, NonNull};
 use std::slice;
 use std::sync::Arc;
 use thiserror::Error;
-use wasmtime_environ::entity::{EntityRef, EntitySet, PrimaryMap};
-use wasmtime_environ::wasm::{
-    DefinedFuncIndex, DefinedMemoryIndex, DefinedTableIndex, GlobalInit, SignatureIndex, WasmType,
-};
 use wasmtime_environ::{
-    ir, HostPtr, MemoryInitialization, MemoryInitializer, Module, ModuleType, TableInitializer,
-    VMOffsets, WASM_PAGE_SIZE,
+    DefinedFuncIndex, DefinedMemoryIndex, DefinedTableIndex, EntityRef, EntitySet, FunctionInfo,
+    GlobalInit, HostPtr, MemoryInitialization, MemoryInitializer, Module, ModuleType, PrimaryMap,
+    SignatureIndex, TableInitializer, TrapCode, VMOffsets, WasmType, WASM_PAGE_SIZE,
 };
 
 mod pooling;
@@ -37,8 +34,12 @@ pub struct InstanceAllocationRequest<'a> {
     /// The module being instantiated.
     pub module: Arc<Module>,
 
-    /// The finished (JIT) functions for the module.
-    pub finished_functions: &'a PrimaryMap<DefinedFuncIndex, *mut [VMFunctionBody]>,
+    /// The base address of where JIT functions are located.
+    pub image_base: usize,
+
+    /// Descriptors about each compiled function, such as the offset from
+    /// `image_base`.
+    pub functions: &'a PrimaryMap<DefinedFuncIndex, FunctionInfo>,
 
     /// The imports to use for the instantiation.
     pub imports: Imports<'a>,
@@ -64,6 +65,16 @@ pub struct InstanceAllocationRequest<'a> {
     /// We use a number of `PhantomPinned` declarations to indicate this to the
     /// compiler. More info on this in `wasmtime/src/store.rs`
     pub store: Option<*mut dyn Store>,
+
+    /// A list of all wasm data that can be referenced by the module that
+    /// will be allocated. The `Module` given here has active/passive data
+    /// segments that are specified as relative indices into this list of bytes.
+    ///
+    /// Note that this is an unsafe pointer. The pointer is expected to live for
+    /// the entire duration of the instance at this time. It's the
+    /// responsibility of the callee when allocating to ensure that this data
+    /// outlives the instance.
+    pub wasm_data: *const [u8],
 }
 
 /// An link error while instantiating a module.
@@ -279,14 +290,22 @@ fn initialize_tables(instance: &mut Instance, module: &Module) -> Result<(), Ins
 fn get_memory_init_start(
     init: &MemoryInitializer,
     instance: &Instance,
-) -> Result<u32, InstantiationError> {
+) -> Result<u64, InstantiationError> {
     match init.base {
         Some(base) => {
+            let mem64 = instance.module.memory_plans[init.memory_index]
+                .memory
+                .memory64;
             let val = unsafe {
-                if let Some(def_index) = instance.module.defined_global_index(base) {
-                    *instance.global(def_index).as_u32()
+                let global = if let Some(def_index) = instance.module.defined_global_index(base) {
+                    instance.global(def_index)
                 } else {
-                    *(*instance.imported_global(base).from).as_u32()
+                    &*instance.imported_global(base).from
+                };
+                if mem64 {
+                    *global.as_u64()
+                } else {
+                    u64::from(*global.as_u32())
                 }
             };
 
@@ -305,8 +324,9 @@ fn check_memory_init_bounds(
     for init in initializers {
         let memory = instance.get_memory(init.memory_index);
         let start = get_memory_init_start(init, instance)?;
-        let start = usize::try_from(start).unwrap();
-        let end = start.checked_add(init.data.len());
+        let end = usize::try_from(start)
+            .ok()
+            .and_then(|start| start.checked_add(init.data.len()));
 
         match end {
             Some(end) if end <= memory.current_length => {
@@ -331,10 +351,10 @@ fn initialize_memories(
         instance
             .memory_init_segment(
                 init.memory_index,
-                &init.data,
+                init.data.clone(),
                 get_memory_init_start(init, instance)?,
                 0,
-                init.data.len() as u32,
+                init.data.end - init.data.start,
             )
             .map_err(InstantiationError::Trap)?;
     }
@@ -385,13 +405,11 @@ fn initialize_instance(
                 let slice =
                     unsafe { slice::from_raw_parts_mut(memory.base, memory.current_length) };
 
-                for (page_index, page) in pages.iter().enumerate() {
-                    if let Some(data) = page {
-                        debug_assert_eq!(data.len(), WASM_PAGE_SIZE as usize);
-                        let start = page_index * WASM_PAGE_SIZE as usize;
-                        let end = start + WASM_PAGE_SIZE as usize;
-                        slice[start..end].copy_from_slice(data);
-                    }
+                for (page_index, page) in pages {
+                    debug_assert_eq!(page.end - page.start, WASM_PAGE_SIZE);
+                    let start = (*page_index * u64::from(WASM_PAGE_SIZE)) as usize;
+                    let end = start + WASM_PAGE_SIZE as usize;
+                    slice[start..end].copy_from_slice(instance.wasm_data(page.clone()));
                 }
             }
 
@@ -399,7 +417,7 @@ fn initialize_instance(
             // the expected behavior of the bulk memory spec.
             if *out_of_bounds {
                 return Err(InstantiationError::Trap(Trap::wasm(
-                    ir::TrapCode::HeapOutOfBounds,
+                    TrapCode::HeapOutOfBounds,
                 )));
             }
         }
@@ -469,7 +487,8 @@ unsafe fn initialize_vmcontext(instance: &mut Instance, req: InstanceAllocationR
 
         let (func_ptr, vmctx) = if let Some(def_index) = instance.module.defined_func_index(index) {
             (
-                NonNull::new(req.finished_functions[def_index] as *mut _).unwrap(),
+                NonNull::new((req.image_base + req.functions[def_index].start as usize) as *mut _)
+                    .unwrap(),
                 instance.vmctx_ptr(),
             )
         } else {
@@ -524,7 +543,7 @@ unsafe fn initialize_vmcontext_globals(instance: &Instance) {
             GlobalInit::I64Const(x) => *(*to).as_i64_mut() = x,
             GlobalInit::F32Const(x) => *(*to).as_f32_bits_mut() = x,
             GlobalInit::F64Const(x) => *(*to).as_f64_bits_mut() = x,
-            GlobalInit::V128Const(x) => *(*to).as_u128_bits_mut() = x.0,
+            GlobalInit::V128Const(x) => *(*to).as_u128_mut() = x,
             GlobalInit::GetGlobal(x) => {
                 let from = if let Some(def_x) = module.defined_global_index(x) {
                     instance.global(def_x)
@@ -646,8 +665,9 @@ unsafe impl InstanceAllocator for OnDemandInstanceAllocator {
                 memories,
                 tables,
                 dropped_elements: EntitySet::with_capacity(req.module.passive_elements.len()),
-                dropped_data: EntitySet::with_capacity(req.module.passive_data.len()),
+                dropped_data: EntitySet::with_capacity(req.module.passive_data_map.len()),
                 host_state,
+                wasm_data: &*req.wasm_data,
                 vmctx: VMContext {
                     _marker: marker::PhantomPinned,
                 },
