@@ -20,8 +20,8 @@ use std::marker;
 use std::mem;
 use std::sync::{Arc, Mutex};
 use wasmtime_environ::{
-    entity::{EntitySet, PrimaryMap},
-    HostPtr, MemoryStyle, Module, Tunables, VMOffsets, VMOffsetsFields, WASM_PAGE_SIZE,
+    EntitySet, HostPtr, MemoryStyle, Module, PrimaryMap, Tunables, VMOffsets, VMOffsetsFields,
+    WASM_PAGE_SIZE,
 };
 
 cfg_if::cfg_if! {
@@ -89,7 +89,7 @@ pub struct ModuleLimits {
     pub table_elements: u32,
 
     /// The maximum number of pages for any linear memory defined in a module.
-    pub memory_pages: u32,
+    pub memory_pages: u64,
 }
 
 impl ModuleLimits {
@@ -197,7 +197,7 @@ impl ModuleLimits {
                 );
             }
 
-            if let MemoryStyle::Dynamic = plan.style {
+            if let MemoryStyle::Dynamic { .. } = plan.style {
                 bail!(
                     "memory index {} has an unsupported dynamic memory plan style",
                     i,
@@ -364,6 +364,7 @@ impl InstancePool {
                     dropped_elements: EntitySet::new(),
                     dropped_data: EntitySet::new(),
                     host_state: Box::new(()),
+                    wasm_data: &[],
                     vmctx: VMContext {
                         _marker: marker::PhantomPinned,
                     },
@@ -382,6 +383,7 @@ impl InstancePool {
         instance.module = req.module.clone();
         instance.offsets = VMOffsets::new(HostPtr, instance.module.as_ref());
         instance.host_state = std::mem::replace(&mut req.host_state, Box::new(()));
+        instance.wasm_data = &*req.wasm_data;
 
         let mut limiter = req.store.and_then(|s| (*s).limiter());
         Self::set_instance_memories(
@@ -455,7 +457,7 @@ impl InstancePool {
                 .expect("failed to reset guard pages");
             drop(&mut memory); // require mutable on all platforms, not just uffd
 
-            let size = (memory.size() as usize) * (WASM_PAGE_SIZE as usize);
+            let size = memory.byte_size();
             drop(memory);
             decommit_memory_pages(base, size).expect("failed to decommit linear memory pages");
         }
@@ -492,6 +494,7 @@ impl InstancePool {
         // fresh allocation later on.
         instance.module = self.empty_module.clone();
         instance.offsets = VMOffsets::new(HostPtr, &self.empty_module);
+        instance.wasm_data = &[];
 
         self.free_list.lock().unwrap().push(index);
     }
@@ -499,7 +502,7 @@ impl InstancePool {
     fn set_instance_memories(
         instance: &mut Instance,
         mut memories: impl Iterator<Item = *mut u8>,
-        max_pages: u32,
+        max_pages: u64,
         mut limiter: Option<&mut dyn ResourceLimiter>,
     ) -> Result<(), InstantiationError> {
         let module = instance.module.as_ref();
@@ -527,7 +530,6 @@ impl InstancePool {
         }
 
         debug_assert!(instance.dropped_data.is_empty());
-        instance.dropped_data.resize(module.passive_data.len());
 
         Ok(())
     }
@@ -599,7 +601,7 @@ struct MemoryPool {
     initial_memory_offset: usize,
     max_memories: usize,
     max_instances: usize,
-    max_wasm_pages: u32,
+    max_wasm_pages: u64,
 }
 
 impl MemoryPool {
@@ -1000,7 +1002,7 @@ unsafe impl InstanceAllocator for PoolingInstanceAllocator {
                         // If there was an out of bounds access observed in initialization, return a trap
                         if *out_of_bounds {
                             return Err(InstantiationError::Trap(crate::traphandlers::Trap::wasm(
-                                wasmtime_environ::ir::TrapCode::HeapOutOfBounds,
+                                wasmtime_environ::TrapCode::HeapOutOfBounds,
                             )));
                         }
 
@@ -1050,10 +1052,8 @@ mod test {
     use super::*;
     use crate::{Imports, VMSharedSignatureIndex};
     use wasmtime_environ::{
-        entity::EntityRef,
-        ir::Type,
-        wasm::{Global, GlobalInit, Memory, SignatureIndex, Table, TableElementType, WasmType},
-        MemoryPlan, ModuleType, TablePlan, TableStyle,
+        EntityRef, Global, GlobalInit, Memory, MemoryPlan, ModuleType, SignatureIndex, Table,
+        TablePlan, TableStyle, WasmType,
     };
 
     #[test]
@@ -1088,7 +1088,6 @@ mod test {
             style: TableStyle::CallerChecksSignature,
             table: Table {
                 wasm_ty: WasmType::FuncRef,
-                ty: TableElementType::Func,
                 minimum: 0,
                 maximum: None,
             },
@@ -1118,6 +1117,7 @@ mod test {
                 minimum: 0,
                 maximum: None,
                 shared: false,
+                memory64: false,
             },
             pre_guard_size: 0,
             offset_guard_size: 0,
@@ -1143,7 +1143,6 @@ mod test {
 
         module.globals.push(Global {
             wasm_ty: WasmType::I32,
-            ty: Type::int(32).unwrap(),
             mutability: false,
             initializer: GlobalInit::I32Const(0),
         });
@@ -1207,7 +1206,6 @@ mod test {
             style: TableStyle::CallerChecksSignature,
             table: Table {
                 wasm_ty: WasmType::FuncRef,
-                ty: TableElementType::Func,
                 minimum: 0,
                 maximum: None,
             },
@@ -1234,6 +1232,7 @@ mod test {
                 minimum: 0,
                 maximum: None,
                 shared: false,
+                memory64: false,
             },
             pre_guard_size: 0,
             offset_guard_size: 0,
@@ -1256,7 +1255,6 @@ mod test {
 
         module.globals.push(Global {
             wasm_ty: WasmType::I32,
-            ty: Type::int(32).unwrap(),
             mutability: false,
             initializer: GlobalInit::I32Const(0),
         });
@@ -1279,7 +1277,6 @@ mod test {
             style: TableStyle::CallerChecksSignature,
             table: Table {
                 wasm_ty: WasmType::FuncRef,
-                ty: TableElementType::Func,
                 minimum: 11,
                 maximum: None,
             },
@@ -1308,6 +1305,7 @@ mod test {
                 minimum: 6,
                 maximum: None,
                 shared: false,
+                memory64: false,
             },
             pre_guard_size: 0,
             offset_guard_size: 0,
@@ -1328,11 +1326,12 @@ mod test {
 
         let mut module = Module::default();
         module.memory_plans.push(MemoryPlan {
-            style: MemoryStyle::Dynamic,
+            style: MemoryStyle::Dynamic { reserve: 0 },
             memory: Memory {
                 minimum: 1,
                 maximum: None,
                 shared: false,
+                memory64: false,
             },
             offset_guard_size: 0,
             pre_guard_size: 0,
@@ -1394,7 +1393,7 @@ mod test {
 
         let mut handles = Vec::new();
         let module = Arc::new(Module::default());
-        let finished_functions = &PrimaryMap::new();
+        let functions = &PrimaryMap::new();
 
         for _ in (0..3).rev() {
             handles.push(
@@ -1403,7 +1402,8 @@ mod test {
                         PoolingAllocationStrategy::NextAvailable,
                         InstanceAllocationRequest {
                             module: module.clone(),
-                            finished_functions,
+                            image_base: 0,
+                            functions,
                             imports: Imports {
                                 functions: &[],
                                 tables: &[],
@@ -1413,6 +1413,7 @@ mod test {
                             shared_signatures: VMSharedSignatureIndex::default().into(),
                             host_state: Box::new(()),
                             store: None,
+                            wasm_data: &[],
                         },
                     )
                     .expect("allocation should succeed"),
@@ -1425,7 +1426,8 @@ mod test {
             PoolingAllocationStrategy::NextAvailable,
             InstanceAllocationRequest {
                 module: module.clone(),
-                finished_functions,
+                functions,
+                image_base: 0,
                 imports: Imports {
                     functions: &[],
                     tables: &[],
@@ -1435,6 +1437,7 @@ mod test {
                 shared_signatures: VMSharedSignatureIndex::default().into(),
                 host_state: Box::new(()),
                 store: None,
+                wasm_data: &[],
             },
         ) {
             Err(InstantiationError::Limit(3)) => {}
