@@ -57,14 +57,14 @@
 //!   ```
 
 use crate::externref::VMExternRef;
-use crate::table::Table;
+use crate::instance::Instance;
+use crate::table::{Table, TableElementType};
 use crate::traphandlers::{raise_lib_trap, Trap};
 use crate::vmcontext::{VMCallerCheckedAnyfunc, VMContext};
+use backtrace::Backtrace;
 use std::mem;
 use std::ptr::{self, NonNull};
-use wasmtime_environ::wasm::{
-    DataIndex, ElemIndex, GlobalIndex, MemoryIndex, TableElementType, TableIndex,
-};
+use wasmtime_environ::{DataIndex, ElemIndex, GlobalIndex, MemoryIndex, TableIndex, TrapCode};
 
 const TOINT_32: f32 = 1.0 / f32::EPSILON;
 const TOINT_64: f64 = 1.0 / f64::EPSILON;
@@ -187,14 +187,15 @@ pub extern "C" fn wasmtime_f64_nearest(x: f64) -> f64 {
 /// Implementation of memory.grow for locally-defined 32-bit memories.
 pub unsafe extern "C" fn wasmtime_memory32_grow(
     vmctx: *mut VMContext,
-    delta: u32,
+    delta: u64,
     memory_index: u32,
-) -> u32 {
+) -> usize {
     let instance = (*vmctx).instance_mut();
     let memory_index = MemoryIndex::from_u32(memory_index);
-    instance
-        .memory_grow(memory_index, delta)
-        .unwrap_or(u32::max_value())
+    match instance.memory_grow(memory_index, delta) {
+        Some(size_in_bytes) => size_in_bytes / (wasmtime_environ::WASM_PAGE_SIZE as usize),
+        None => usize::max_value(),
+    }
 }
 
 /// Implementation of `table.grow`.
@@ -210,9 +211,7 @@ pub unsafe extern "C" fn wasmtime_table_grow(
     let table_index = TableIndex::from_u32(table_index);
     let element = match instance.table_element_type(table_index) {
         TableElementType::Func => (init_value as *mut VMCallerCheckedAnyfunc).into(),
-        TableElementType::Val(ty) => {
-            debug_assert_eq!(ty, crate::ref_type());
-
+        TableElementType::Extern => {
             let init_value = if init_value.is_null() {
                 None
             } else {
@@ -245,8 +244,7 @@ pub unsafe extern "C" fn wasmtime_table_fill(
                 let val = val as *mut VMCallerCheckedAnyfunc;
                 table.fill(dst, val.into(), len)
             }
-            TableElementType::Val(ty) => {
-                debug_assert_eq!(ty, crate::ref_type());
+            TableElementType::Extern => {
                 let val = if val.is_null() {
                     None
                 } else {
@@ -314,10 +312,10 @@ pub unsafe extern "C" fn wasmtime_elem_drop(vmctx: *mut VMContext, elem_index: u
 pub unsafe extern "C" fn wasmtime_memory_copy(
     vmctx: *mut VMContext,
     dst_index: u32,
-    dst: u32,
+    dst: u64,
     src_index: u32,
-    src: u32,
-    len: u32,
+    src: u64,
+    len: u64,
 ) {
     let result = {
         let src_index = MemoryIndex::from_u32(src_index);
@@ -334,14 +332,14 @@ pub unsafe extern "C" fn wasmtime_memory_copy(
 pub unsafe extern "C" fn wasmtime_memory_fill(
     vmctx: *mut VMContext,
     memory_index: u32,
-    dst: u32,
+    dst: u64,
     val: u32,
-    len: u32,
+    len: u64,
 ) {
     let result = {
         let memory_index = MemoryIndex::from_u32(memory_index);
         let instance = (*vmctx).instance_mut();
-        instance.memory_fill(memory_index, dst, val, len)
+        instance.memory_fill(memory_index, dst, val as u8, len)
     };
     if let Err(trap) = result {
         raise_lib_trap(trap);
@@ -353,7 +351,7 @@ pub unsafe extern "C" fn wasmtime_memory_init(
     vmctx: *mut VMContext,
     memory_index: u32,
     data_index: u32,
-    dst: u32,
+    dst: u64,
     src: u32,
     len: u32,
 ) {
@@ -449,40 +447,103 @@ impl std::fmt::Display for Unimplemented {
 
 /// Implementation of `memory.atomic.notify` for locally defined memories.
 pub unsafe extern "C" fn wasmtime_memory_atomic_notify(
-    _vmctx: *mut VMContext,
-    _memory_index: u32,
-    _addr: u32,
+    vmctx: *mut VMContext,
+    memory_index: u32,
+    addr: usize,
     _count: u32,
 ) -> u32 {
-    raise_lib_trap(Trap::User(Box::new(Unimplemented(
-        "wasm atomics (fn wasmtime_memory_atomic_notify) unsupported",
-    ))));
+    let result = {
+        let memory = MemoryIndex::from_u32(memory_index);
+        let instance = (*vmctx).instance();
+        // this should never overflow since addr + 4 either hits a guard page
+        // or it's been validated to be in-bounds already. Double-check for now
+        // just to be sure.
+        let addr_to_check = addr.checked_add(4).unwrap();
+        validate_atomic_addr(instance, memory, addr_to_check).and_then(|()| {
+            Err(Trap::User(Box::new(Unimplemented(
+                "wasm atomics (fn wasmtime_memory_atomic_notify) unsupported",
+            ))))
+        })
+    };
+    match result {
+        Ok(n) => n,
+        Err(e) => raise_lib_trap(e),
+    }
 }
 
 /// Implementation of `memory.atomic.wait32` for locally defined memories.
 pub unsafe extern "C" fn wasmtime_memory_atomic_wait32(
-    _vmctx: *mut VMContext,
-    _memory_index: u32,
-    _addr: u32,
+    vmctx: *mut VMContext,
+    memory_index: u32,
+    addr: usize,
     _expected: u32,
     _timeout: u64,
 ) -> u32 {
-    raise_lib_trap(Trap::User(Box::new(Unimplemented(
-        "wasm atomics (fn wasmtime_memory_atomic_wait32) unsupported",
-    ))));
+    let result = {
+        let memory = MemoryIndex::from_u32(memory_index);
+        let instance = (*vmctx).instance();
+        // see wasmtime_memory_atomic_notify for why this shouldn't overflow
+        // but we still double-check
+        let addr_to_check = addr.checked_add(4).unwrap();
+        validate_atomic_addr(instance, memory, addr_to_check).and_then(|()| {
+            Err(Trap::User(Box::new(Unimplemented(
+                "wasm atomics (fn wasmtime_memory_atomic_wait32) unsupported",
+            ))))
+        })
+    };
+    match result {
+        Ok(n) => n,
+        Err(e) => raise_lib_trap(e),
+    }
 }
 
 /// Implementation of `memory.atomic.wait64` for locally defined memories.
 pub unsafe extern "C" fn wasmtime_memory_atomic_wait64(
-    _vmctx: *mut VMContext,
-    _memory_index: u32,
-    _addr: u32,
+    vmctx: *mut VMContext,
+    memory_index: u32,
+    addr: usize,
     _expected: u64,
     _timeout: u64,
 ) -> u32 {
-    raise_lib_trap(Trap::User(Box::new(Unimplemented(
-        "wasm atomics (fn wasmtime_memory_atomic_wait32) unsupported",
-    ))));
+    let result = {
+        let memory = MemoryIndex::from_u32(memory_index);
+        let instance = (*vmctx).instance();
+        // see wasmtime_memory_atomic_notify for why this shouldn't overflow
+        // but we still double-check
+        let addr_to_check = addr.checked_add(8).unwrap();
+        validate_atomic_addr(instance, memory, addr_to_check).and_then(|()| {
+            Err(Trap::User(Box::new(Unimplemented(
+                "wasm atomics (fn wasmtime_memory_atomic_wait64) unsupported",
+            ))))
+        })
+    };
+    match result {
+        Ok(n) => n,
+        Err(e) => raise_lib_trap(e),
+    }
+}
+
+/// For atomic operations we still check the actual address despite this also
+/// being checked via the `heap_addr` instruction in cranelift. The reason for
+/// that is because the `heap_addr` instruction can defer to a later segfault to
+/// actually recognize the out-of-bounds whereas once we're running Rust code
+/// here we don't want to segfault.
+///
+/// In the situations where bounds checks were elided in JIT code (because oob
+/// would then be later guaranteed to segfault) this manual check is here
+/// so we don't segfault from Rust.
+unsafe fn validate_atomic_addr(
+    instance: &Instance,
+    memory: MemoryIndex,
+    addr: usize,
+) -> Result<(), Trap> {
+    if addr > instance.get_memory(memory).current_length {
+        return Err(Trap::Wasm {
+            trap_code: TrapCode::HeapOutOfBounds,
+            backtrace: Backtrace::new_unresolved(),
+        });
+    }
+    Ok(())
 }
 
 /// Hook for when an instance runs out of fuel.
