@@ -159,6 +159,10 @@ mod test_vmmemory_import {
 
 /// The fields compiled code needs to access to utilize a WebAssembly global
 /// variable imported from another instance.
+///
+/// Note that unlike with functions, tables, and memories, `VMGlobalImport`
+/// doesn't include a `vmctx` pointer. Globals are never resized, and don't
+/// require a `vmctx` pointer to access.
 #[derive(Debug, Copy, Clone)]
 #[repr(C)]
 pub struct VMGlobalImport {
@@ -597,12 +601,6 @@ impl VMBuiltinFunctionsArray {
 
         ptrs[BuiltinFunctionIndex::memory32_grow().index() as usize] =
             wasmtime_memory32_grow as usize;
-        ptrs[BuiltinFunctionIndex::imported_memory32_grow().index() as usize] =
-            wasmtime_imported_memory32_grow as usize;
-        ptrs[BuiltinFunctionIndex::memory32_size().index() as usize] =
-            wasmtime_memory32_size as usize;
-        ptrs[BuiltinFunctionIndex::imported_memory32_size().index() as usize] =
-            wasmtime_imported_memory32_size as usize;
         ptrs[BuiltinFunctionIndex::table_copy().index() as usize] = wasmtime_table_copy as usize;
         ptrs[BuiltinFunctionIndex::table_grow_funcref().index() as usize] =
             wasmtime_table_grow as usize;
@@ -612,8 +610,6 @@ impl VMBuiltinFunctionsArray {
         ptrs[BuiltinFunctionIndex::elem_drop().index() as usize] = wasmtime_elem_drop as usize;
         ptrs[BuiltinFunctionIndex::memory_copy().index() as usize] = wasmtime_memory_copy as usize;
         ptrs[BuiltinFunctionIndex::memory_fill().index() as usize] = wasmtime_memory_fill as usize;
-        ptrs[BuiltinFunctionIndex::imported_memory_fill().index() as usize] =
-            wasmtime_imported_memory_fill as usize;
         ptrs[BuiltinFunctionIndex::memory_init().index() as usize] = wasmtime_memory_init as usize;
         ptrs[BuiltinFunctionIndex::data_drop().index() as usize] = wasmtime_data_drop as usize;
         ptrs[BuiltinFunctionIndex::drop_externref().index() as usize] =
@@ -630,16 +626,10 @@ impl VMBuiltinFunctionsArray {
             wasmtime_table_fill as usize;
         ptrs[BuiltinFunctionIndex::memory_atomic_notify().index() as usize] =
             wasmtime_memory_atomic_notify as usize;
-        ptrs[BuiltinFunctionIndex::imported_memory_atomic_notify().index() as usize] =
-            wasmtime_imported_memory_atomic_notify as usize;
         ptrs[BuiltinFunctionIndex::memory_atomic_wait32().index() as usize] =
             wasmtime_memory_atomic_wait32 as usize;
-        ptrs[BuiltinFunctionIndex::imported_memory_atomic_wait32().index() as usize] =
-            wasmtime_imported_memory_atomic_wait32 as usize;
         ptrs[BuiltinFunctionIndex::memory_atomic_wait64().index() as usize] =
             wasmtime_memory_atomic_wait64 as usize;
-        ptrs[BuiltinFunctionIndex::imported_memory_atomic_wait64().index() as usize] =
-            wasmtime_imported_memory_atomic_wait64 as usize;
         ptrs[BuiltinFunctionIndex::out_of_gas().index() as usize] = wasmtime_out_of_gas as usize;
 
         if cfg!(debug_assertions) {
@@ -758,8 +748,6 @@ mod test_vminterrupts {
 /// The struct here is empty, as the sizes of these fields are dynamic, and
 /// we can't describe them in Rust's type system. Sufficient memory is
 /// allocated at runtime.
-///
-/// TODO: We could move the globals into the `vmctx` allocation too.
 #[derive(Debug)]
 #[repr(C, align(16))] // align 16 since globals are aligned to that and contained inside
 pub struct VMContext {
